# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
import psutil
import signal
import sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
os.environ['GLOG_v'] = '3'
import argparse
import random
import numpy as np
import subprocess
import argparse
import mindspore as ms
import mindspore.communication.management as D
from mindspore import context
from loguru import logger

from MindsporeTrainer.utils import *
from MindsporeTrainer.task import load_tasks, get_task
from MindsporeTrainer.training import DistributedTrainer
from MindsporeTrainer.training import get_args as get_training_args
from MindsporeTrainer.optims import get_args as get_optims_args
from MindsporeTrainer.modeling.modeling_adapter import *


def kill_children(proc=None, recursive = True):
    if proc is None:
        proc = psutil.Process()
    _children = proc.children(recursive=False)
    for c in _children:
        try:
            if recursive:
                kill_children(c, recursive=recursive)
            os.kill(c.pid, signal.SIGKILL)
        except:
            pass

        for c in _children:
            try:
                c.wait(1)
            except:
                pass

class LoadTaskAction(argparse.Action):
    _registered = False
    def __call__(self, parser, args, values, option_string=None):
        setattr(args, self.dest, values)
        if not self._registered:
            load_tasks(args.task_dir, values)
            all_tasks = get_task()
            if values=="*":
                for task in all_tasks.values():
                    parser.add_argument_group(title=f'Task {task._meta["name"]}', description=task._meta["desc"])
                return

            # assert values.lower() in all_tasks, f'{values} is not registed. Valid tasks {list(all_tasks.keys())}'
            task = get_task(values)
            group = parser.add_argument_group(title=f'Task {task._meta["name"]}', description=task._meta["desc"])
            task.add_arguments(group)
            type(self)._registered = True

def build_argument_parser():
    parser = argparse.ArgumentParser(parents=[get_optims_args(), get_training_args()], formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    ## Required parameters
    parser.add_argument("--task_dir",
                default=None,
                type=str,
                required=False,
                help="The directory to load customized tasks.")
    parser.add_argument("--task_name",
                default=None,
                type=str,
                action=LoadTaskAction,
                required=True,
                help="The name of the task to train. To list all registered tasks, use \"*\" as the name, e.g. \n"
                "\npython -m DeBERTa.apps.run --task_name \"*\" --help")

    parser.add_argument("--data_dir",
                default=None,
                type=str,
                required=False,
                help="The input data dir. Should contain the .tsv files (or other data files) for the task.")

    parser.add_argument("--output_dir",
                default=None,
                type=str,
                required=True,
                help="The output directory where the model checkpoints will be written.")

    parser.add_argument("--do_train",
                default=False,
                action='store_true',
                help="Whether to run training.")

    parser.add_argument("--do_eval",
                default=False,
                action='store_true',
                help="Whether to run eval on the dev set.")

    parser.add_argument("--do_predict",
                default=False,
                action='store_true',
                help="Whether to run prediction on the test set.")

    parser.add_argument("--eval_batch_size",
                default=32,
                type=int,
                help="Total batch size for eval.")

    parser.add_argument("--predict_batch_size",
                default=32,
                type=int,
                help="Total batch size for prediction.")

    parser.add_argument('--tag',
                type=str,
                default='final',
                help="The tag name of current prediction/runs.")

    parser.add_argument('--debug',
                default=False,
                action='store_true',
                help="Whether to cache cooked binary features")

    parser.add_argument("--num_train_epochs",
                default=1,
                type=int,
                help="Total number of training epochs to perform.")

    parser.add_argument('--device_target', 
                        type=str, 
                        default='Ascend', 
                        choices=['Ascend', 'GPU'],
                        help='device where the code will be implemented. (Default: Ascend)')
    parser.add_argument('--run_mode', 
                        type=str, 
                        default='GRAPH', 
                        choices=['GRAPH', 'PY'],
                        help='0: GRAPH_MODE, 1: PY_NATIVE_MODE')
    parser.add_argument("--distribute", 
                        default=False, 
                        action='store_true',
                        help="Run distribute, default is false.")

    parser.add_argument("--device_id", type=str, default="0", help="Device id, default is 0.")
    parser.add_argument("--device_num", type=int, default=1, help="Use device nums, default is 1.")
    parser.add_argument("--enable_save_ckpt", type=str, default="true", choices=["true", "false"],
                        help="Enable save checkpoint, default is true.")
    parser.add_argument("--enable_lossscale", type=str, default="true", choices=["true", "false"],
                        help="Use lossscale or not, default is not.")
    parser.add_argument("--do_shuffle", type=str, default="true", choices=["true", "false"],
                        help="Enable shuffle for dataset, default is true.")
    parser.add_argument("--enable_data_sink",
                        default=False,
                        action='store_true',
                        help="Enable data sink, default is false.")
    parser.add_argument("--data_sink_steps", type=int, default="1", help="Sink steps for each epoch, default is 1.")
    parser.add_argument("--accumulation_steps", type=int, default="1",
                        help="Accumulating gradients N times before weight update, default is 1.")
    parser.add_argument("--allreduce_post_accumulation", type=str, default="true", choices=["true", "false"],
                        help="Whether to allreduce after accumulation of N steps or after each step, default is true.")
    # parser.add_argument("--save_checkpoint_path", type=str, default="", help="Save checkpoint path")
    parser.add_argument("--load_checkpoint_path", type=str, default="", help="Load checkpoint file path")
    parser.add_argument("--restore_by_prefix", default=False, action='store_true', help="Load all of the parameters")
    parser.add_argument("--prefix", type=str, default="backbone", help="remove the prefix when load ckpt")
    parser.add_argument("--save_eval_steps", type=int, default=1000, help="Save checkpoint and evaluate steps, "
                                                                                "default is 1000.")
    parser.add_argument("--train_steps", type=int, default=-1, help="Training Steps, default is -1, "
                                                                    "meaning run all steps according to epoch number.")
    parser.add_argument("--save_checkpoint_num", type=int, default=1, help="Save checkpoint numbers, default is 1.")

    # parser.add_argument("--schema_dir", type=str, default="", help="Schema path, it is better to use absolute path")
    parser.add_argument("--enable_graph_kernel", type=str, default="auto", choices=["auto", "true", "false"],
                        help="Accelerate by graph kernel, default is auto.")
    parser.add_argument("--save_graphs", 
                        default=False,
                        action='store_true',
                        help="Whether to save graphs")
    parser.add_argument("--thor", 
                        default=False,
                        action='store_true',
                        help="Whether to convert model to thor optimizer")
    return parser


def main(args):
    if not args.do_train and not args.do_eval and not args.do_predict:
        raise ValueError("At least one of `do_train` or `do_eval` or `do_predict` must be True.")
    random.seed(args.seed)
    np.random.seed(args.seed)
    if args.accumulation_steps > 1:
        logger.info("accumulation steps: {}".format(args.accumulation_steps))
        logger.info("global batch size: {}".format(args.train_batch_size))
        if args.enable_data_sink == "true":
            args.data_sink_steps *= args.accumulation_steps
            logger.info("data sink steps: {}".format(args.data_sink_steps))
        if args.enable_save_ckpt == "true":
            args.save_eval_steps *= args.accumulation_steps
            logger.info("save checkpoint steps: {}".format(args.save_eval_steps))
        args.train_batch_size = args.train_batch_size // args.accumulation_steps
        if args.train_steps > 0:
            args.train_steps = args.train_steps * args.accumulation_steps

    task = get_task(args.task_name)(args)
    trainer = DistributedTrainer(args, task)
    trainer.run()


def set_context(args):
    """set_parameter"""
    target = args.device_target
    if target == "CPU":
        args.distribute = False
    ms.context.set_context(reserve_class_name_in_scope=False)
    # init context
    if args.run_mode == 'GRAPH':
        if target == "Ascend":
            rank_save_graphs_path = os.path.join(args.output_dir, "soma", str(os.getenv('DEVICE_ID')))
            ms.context.set_context(mode=ms.context.GRAPH_MODE, 
                                   device_target=target, 
                                   save_graphs=args.save_graphs,
                                   save_graphs_path=args.output_dir, 
                                   device_id=int(args.device_id))
        else:
            ms.context.set_context(mode=ms.context.GRAPH_MODE, 
                                   device_target=target, 
                                   save_graphs=args.save_graphs, 
                                   save_graphs_path=args.output_dir, 
                                   device_id=int(args.device_id))

    else:
        ms.context.set_context(mode=ms.context.PYNATIVE_MODE, 
                               device_target=target, 
                               save_graphs=args.save_graphs, 
                               save_graphs_path=args.output_dir, 
                               device_id=int(args.device_id))
    if args.enable_graph_kernel == "true":
        if args.device_target == 'GPU':
            context.set_context(enable_graph_kernel=True)
        else:
            logger.warning('Graph kernel only supports GPU back-end now, run with graph kernel off.')
    if args.distribute:
        if target == "Ascend":
            context.reset_auto_parallel_context()
            ms.context.set_auto_parallel_context(device_num=args.device_num, 
                                                 parallel_mode=ms.context.ParallelMode.DATA_PARALLEL,
                                                 gradients_mean=False)

            D.init()
        # GPU target
        else:
            D.init()
            context.reset_auto_parallel_context()
            ms.context.set_auto_parallel_context(device_num=D.get_device_num(),
                                                parallel_mode=ms.context.ParallelMode.DATA_PARALLEL,
                                                gradients_mean=False)


def launch():
    argv = sys.argv

    parser = build_argument_parser()
    parser.parse_known_args()

    args_opt = parser.parse_args()
    os.makedirs(args_opt.output_dir, exist_ok=True)
    device_num = int(os.getenv('RANK_SIZE', '1'))
    args_opt.device_num = device_num
    logger.info(args_opt)
    if len(args_opt.device_id) > 1:
        device_id = args_opt.device_id.split(',')
        device_id = [int(id) for id in device_id]
        if len(device_id) != device_num:
            if device_num < 8:
                raise f'device id: {device_id} should equal to rank size{device_num}'
        args_opt.device_id = device_id
    else:
        # 若device id 只有一个，rank size > 1，那么device id 作为起始设备编号
        device_id = int(args_opt.device_id)
        device_id = [device_id + i for i in range(device_num)]

    if args_opt.distribute:
        if args_opt.device_target == 'Ascend':
            child_process = []
            if args_opt.rank == -1:
                try:
                    for i, id in enumerate(device_id):
                        rank_id = i + args_opt.local_rank
                        os.putenv('RANK_ID', str(rank_id))
                        os.putenv('DEVICE_ID', str(id))
                        child = subprocess.Popen(['python'] + argv + [f'--rank={rank_id}'] + [f'--device_id={id}'])
                        child_process.append(child)
                    for c in child_process:
                        c.wait()
                except Exception as e:
                    try:
                        for c in child_process:
                            c.kill()
                        logger.exception(f'Uncatched exception happened during execution.')
                        import atexit
                        atexit._run_exitfuncs()
                    except:
                        pass
            else:
                set_context(args_opt)
                try:
                    main(args_opt)
                except Exception as ex:
                    try:
                        logger.exception(f'Uncatched exception happened during execution.')
                        import atexit
                        atexit._run_exitfuncs()
                    except:
                        pass
                    kill_children()
                    os._exit(-1)

        else:
            os.environ["MS_WORKER_NUM"] == device_num
            child_process = []
            if args_opt.rank == -1:
                os.putenv('MS_ROLE', "MS_WORKER")
                for id in device_id:
                    rank_id = id + args_opt.local_rank
                    os.putenv('RANK_ID', str(rank_id))
                    os.putenv('DEVICE_ID', str(id))
                    child = subprocess.Popen(['python'] + argv + [f'--rank={rank_id}'] + [f'--device_id={id}'])
                    child_process.append(child)
            else:
                id = 0
            if args_opt.rank == -1:
                os.environ["MS_ROLE"] = "MS_SCHED"
            set_context(args_opt)
            try:
                main(args_opt)
            except Exception as ex:
                try:
                    logger.exception(f'Uncatched exception happened during execution.')
                    import atexit
                    atexit._run_exitfuncs()
                except:
                    pass
                kill_children()
                os._exit(-1)

    else:
        args_opt.rank = 0
        args_opt.local_rank = 0
        args_opt.device_id = device_id[0]
        set_context(args_opt)
        main(args_opt)




# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import Sequence
import numpy as np

from mindspore.common import dtype as mstype
from mindspore.ops import operations as P
from mindspore.ops import composite as C
from mindspore.ops import functional as F
from mindspore.common.tensor import Tensor
from mindspore._checkparam import Validator as validator
from mindspore._checkparam import Rel
from mindspore.nn.optim.optimizer import Optimizer

_adam_opt = C.MultitypeFuncGraph("adam_opt")
_scaler_one = Tensor(1, mstype.int32)
_scaler_ten = Tensor(10, mstype.float32)

@_adam_opt.register("Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor",
                    "Tensor", "Bool", "Bool")
def _update_run_kernel(beta1, beta2, eps, lr, weight_decay, param, m, v, gradient, decay_flags, optim_filter):
    """
    Update parameters by AdamWeightDecay op.
    """
    if optim_filter:
        adam = P.AdamWeightDecay()
        if decay_flags:
            next_param = adam(param, m, v, lr, beta1, beta2, eps, Tensor(weight_decay, mstype.float32), gradient)
        else:
            next_param = adam(param, m, v, lr, beta1, beta2, eps, Tensor(0.0, mstype.float32), gradient)
        return next_param
    return gradient

@_adam_opt.register("Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor",
                    "Tensor", "Bool", "Bool")
def _update_run_op(beta1, beta2, eps, lr, overflow, weight_decay, param, m, v, gradient, decay_flag, optim_filter):
    """
    Update parameters.

    Args:
        beta1 (Tensor): The exponential decay rate for the 1st moment estimations. Should be in range (0.0, 1.0).
        beta2 (Tensor): The exponential decay rate for the 2nd moment estimations. Should be in range (0.0, 1.0).
        eps (Tensor): Term added to the denominator to improve numerical stability. Should be greater than 0.
        lr (Tensor): Learning rate.
        overflow (Tensor): Whether overflow occurs.
        weight_decay (Number): Weight decay. Should be equal to or greater than 0.
        param (Tensor): Parameters.
        m (Tensor): m value of parameters.
        v (Tensor): v value of parameters.
        gradient (Tensor): Gradient of parameters.
        decay_flag (bool): Applies weight decay or not.
        optim_filter (bool): Applies parameter update or not.

    Returns:
        Tensor, the new value of v after updating.
    """
    if optim_filter:
        op_mul = P.Mul()
        op_square = P.Square()
        op_sqrt = P.Sqrt()
        op_cast = P.Cast()
        op_reshape = P.Reshape()
        op_shape = P.Shape()
        op_select = P.Select()

        param_fp32 = op_cast(param, mstype.float32)
        m_fp32 = op_cast(m, mstype.float32)
        v_fp32 = op_cast(v, mstype.float32)
        gradient_fp32 = op_cast(gradient, mstype.float32)

        cond = op_cast(F.fill(mstype.int32, op_shape(m_fp32), 1) * op_reshape(overflow, (())), mstype.bool_)
        next_m = op_mul(beta1, m_fp32) + op_select(cond, m_fp32,\
                op_mul(op_cast(F.tuple_to_array((1.0,)), mstype.float32) - beta1, gradient_fp32))

        next_v = op_mul(beta2, v_fp32) + op_select(cond, v_fp32,\
                op_mul(op_cast(F.tuple_to_array((1.0,)), mstype.float32) - beta2, op_square(gradient_fp32)))

        update = next_m / (eps + op_sqrt(next_v))
        if decay_flag:
            update = op_mul(weight_decay, param_fp32) + update

        update_with_lr = op_mul(lr, update)
        zeros = F.fill(mstype.float32, op_shape(param_fp32), 0)
        next_param = param_fp32 - op_select(cond, zeros, op_reshape(update_with_lr, op_shape(param_fp32)))

        next_param = F.depend(next_param, F.assign(param, op_cast(next_param, F.dtype(param))))
        next_param = F.depend(next_param, F.assign(m, op_cast(next_m, F.dtype(m))))
        next_param = F.depend(next_param, F.assign(v, op_cast(next_v, F.dtype(v))))

        return op_cast(next_param, F.dtype(param))
    return gradient


@_adam_opt.register("Function", "Function", "Function", "Function", "Bool", "Bool", "Bool", "Tensor", "Tensor",
                    "Tensor", "Tensor", "Tensor", "Tensor", "RowTensor", "Tensor", "Tensor", "Tensor", "Bool", "Bool")
def _run_opt_with_sparse(opt, sparse_opt, push, pull, use_locking, use_nesterov, target, beta1_power,
                         beta2_power, beta1, beta2, eps, lr, gradient, param, m, v, ps_parameter, cache_enable):
    """Apply sparse adam optimizer to the weight parameter when the gradient is sparse."""
    success = True
    indices = gradient.indices
    values = gradient.values
    if ps_parameter and not cache_enable:
        op_shape = P.Shape()
        shapes = (op_shape(param), op_shape(m), op_shape(v),
                  op_shape(beta1_power), op_shape(beta2_power), op_shape(lr), op_shape(beta1),
                  op_shape(beta2), op_shape(eps), op_shape(values), op_shape(indices))
        success = F.depend(success, pull(push((beta1_power, beta2_power, lr, beta1, beta2,
                                               eps, values, indices), shapes), param))
        return success

    if not target:
        success = F.depend(success, sparse_opt(param, m, v, beta1_power, beta2_power, lr, beta1, beta2,
                                               eps, values, indices))
    else:
        op_mul = P.Mul()
        op_square = P.Square()
        op_sqrt = P.Sqrt()
        scatter_add = P.ScatterAdd(use_locking)

        success = F.depend(success, F.assign(m, op_mul(beta1, m)))
        success = F.depend(success, F.assign(v, op_mul(beta2, v)))

        grad_indices = gradient.indices
        grad_value = gradient.values

        next_m = scatter_add(m,
                             grad_indices,
                             op_mul(F.tuple_to_array((1.0,)) - beta1, grad_value))

        next_v = scatter_add(v,
                             grad_indices,
                             op_mul(F.tuple_to_array((1.0,)) - beta2, op_square(grad_value)))

        if use_nesterov:
            m_temp = next_m * _scaler_ten
            F.assign(m, op_mul(beta1, next_m))
            div_value = scatter_add(m,
                                    op_mul(grad_indices, _scaler_one),
                                    op_mul(F.tuple_to_array((1.0,)) - beta1, grad_value))
            param_update = div_value / (op_sqrt(next_v) + eps)
            F.assign(m, m_temp / _scaler_ten)
        else:
            param_update = next_m / (op_sqrt(next_v) + eps)

        lr_t = lr * op_sqrt(1 - beta2_power) / (1 - beta1_power)
        next_param = param - lr_t * param_update

        success = F.depend(success, F.assign(param, next_param))
        success = F.depend(success, F.assign(m, next_m))
        success = F.depend(success, F.assign(v, next_v))

    return success


@_adam_opt.register("Function", "Function", "Function", "Function", "Bool", "Bool", "Bool", "Tensor", "Tensor",
                    "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Bool", "Bool")
def _run_opt_with_one_number(opt, sparse_opt, push, pull, use_locking, use_nesterov, target,
                             beta1_power, beta2_power, beta1, beta2, eps, lr, gradient, param,
                             moment1, moment2, ps_parameter, cache_enable):
    """Apply adam optimizer to the weight parameter using Tensor."""
    success = True
    if ps_parameter and not cache_enable:
        op_shape = P.Shape()
        success = F.depend(success, pull(push((beta1_power, beta2_power, lr, beta1, beta2, eps, gradient),
                                              (op_shape(param), op_shape(moment1), op_shape(moment2))), param))
    else:
        success = F.depend(success, opt(param, moment1, moment2, beta1_power, beta2_power, lr, beta1, beta2,
                                        eps, gradient))
    return success


@_adam_opt.register("Function", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor", "Tensor",
                    "Tensor", "Tensor")
def _run_off_load_opt(opt, beta1_power, beta2_power, beta1, beta2, eps, lr, gradient, param, moment1, moment2):
    """Apply AdamOffload optimizer to the weight parameter using Tensor."""
    success = True
    delat_param = opt(moment1, moment2, beta1_power, beta2_power, lr, beta1, beta2, eps, gradient)
    success = F.depend(success, F.assign_add(param, delat_param))
    return success


def _check_param_value(beta1, beta2, eps, prim_name):
    """Check the type of inputs."""
    validator.check_value_type("beta1", beta1, [float], prim_name)
    validator.check_value_type("beta2", beta2, [float], prim_name)
    validator.check_value_type("eps", eps, [float], prim_name)
    validator.check_float_range(beta1, 0.0, 1.0, Rel.INC_NEITHER, "beta1", prim_name)
    validator.check_float_range(beta2, 0.0, 1.0, Rel.INC_NEITHER, "beta2", prim_name)
    validator.check_positive_float(eps, "eps", prim_name)

class AdamWeightDecayX(Optimizer):
    """
    Implements the Adam algorithm to fix the weight decay.

    Note:
        When separating parameter groups, the weight decay in each group will be applied on the parameters if the
        weight decay is positive. When not separating parameter groups, the `weight_decay` in the API will be applied
        on the parameters without 'beta' or 'gamma' in their names if `weight_decay` is positive.

        To improve parameter groups performance, the customized order of parameters can be supported.

    Args:
        params (Union[list[Parameter], list[dict]]): When the `params` is a list of `Parameter` which will be updated,
            the element in `params` must be class `Parameter`. When the `params` is a list of `dict`, the "params",
            "lr", "weight_decay" and "order_params" are the keys can be parsed.

            - params: Required. The value must be a list of `Parameter`.

            - lr: Optional. If "lr" is in the keys, the value of the corresponding learning rate will be used.
              If not, the `learning_rate` in the API will be used.

            - weight_decay: Optional. If "weight_decay" is in the keys, the value of the corresponding weight decay
              will be used. If not, the `weight_decay` in the API will be used.

            - order_params: Optional. If "order_params" is in the keys, the value must be the order of parameters and
              the order will be followed in the optimizer. There are no other keys in the `dict` and the parameters
              which in the 'order_params' must be in one of group parameters.

        learning_rate (Union[float, Tensor, Iterable, LearningRateSchedule]): A value or a graph for the learning rate.
            When the learning_rate is an Iterable or a Tensor in a 1D dimension, use the dynamic learning rate, then
            the i-th step will take the i-th value as the learning rate. When the learning_rate is LearningRateSchedule,
            use dynamic learning rate, the i-th learning rate will be calculated during the process of training
            according to the formula of LearningRateSchedule. When the learning_rate is a float or a Tensor in a zero
            dimension, use fixed learning rate. Other cases are not supported. The float learning rate must be
            equal to or greater than 0. If the type of `learning_rate` is int, it will be converted to float.
            Default: 1e-3.
        beta1 (float): The exponential decay rate for the 1st moment estimations. Default: 0.9.
            Should be in range (0.0, 1.0).
        beta2 (float): The exponential decay rate for the 2nd moment estimations. Default: 0.999.
            Should be in range (0.0, 1.0).
        eps (float): Term added to the denominator to improve numerical stability. Default: 1e-6.
            Should be greater than 0.
        weight_decay (float): Weight decay (L2 penalty). It must be equal to or greater than 0. Default: 0.0.

    Inputs:
        - **gradients** (tuple[Tensor]) - The gradients of `params`, the shape is the same as `params`.
        - **overflow** (tuple[Tensor]) - The overflow flag in dynamiclossscale.

    Outputs:
        tuple[bool], all elements are True.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
        >>> net = Net()
        >>> #1) All parameters use the same learning rate and weight decay
        >>> optim = AdamWeightDecay(params=net.trainable_params())
        >>>
        >>> #2) Use parameter groups and set different values
        >>> conv_params = list(filter(lambda x: 'conv' in x.name, net.trainable_params()))
        >>> no_conv_params = list(filter(lambda x: 'conv' not in x.name, net.trainable_params()))
        >>> group_params = [{'params': conv_params, 'weight_decay': 0.01},
        ...                 {'params': no_conv_params, 'lr': 0.01},
        ...                 {'order_params': net.trainable_params()}]
        >>> optim = AdamWeightDecay(group_params, learning_rate=0.1, weight_decay=0.0)
        >>> # The conv_params's parameters will use default learning rate of 0.1 and weight decay of 0.01.
        >>> # The no_conv_params's parameters will use learning rate of 0.01 and default weight decay of 0.0.
        >>> # The final parameters order in which the optimizer will be followed is the value of 'order_params'.
        >>>
        >>> loss = nn.SoftmaxCrossEntropyWithLogits()
        >>> model = Model(net, loss_fn=loss, optimizer=optim)
   """
    def __init__(self, params, learning_rate=1e-3, beta1=0.9, beta2=0.999, eps=1e-6, weight_decay=0.0):
        super(AdamWeightDecayX, self).__init__(learning_rate, params, weight_decay)
        _check_param_value(beta1, beta2, eps, self.cls_name)
        self.beta1 = Tensor(np.array([beta1]).astype(np.float32))
        self.beta2 = Tensor(np.array([beta2]).astype(np.float32))
        self.eps = Tensor(np.array([eps]).astype(np.float32))
        self.moments1 = self.parameters.clone(prefix="adam_m", init='zeros')
        self.moments2 = self.parameters.clone(prefix="adam_v", init='zeros')
        self.hyper_map = C.HyperMap()
        self.op_select = P.Select()
        self.op_cast = P.Cast()
        self.op_reshape = P.Reshape()
        self.op_shape = P.Shape()
        if isinstance(self.weight_decay, Sequence):
            self.weight_decay = self.weight_decay_tensor_tuple
        else:
            self.weight_decay = Tensor(self.weight_decay)

    def construct(self, gradients, overflow):
        """AdamWeightDecayX"""
        lr = self.get_lr()
        cond = self.op_cast(F.fill(mstype.int32, self.op_shape(self.beta1), 1) *\
                            self.op_reshape(overflow, (())), mstype.bool_)
        beta1 = self.op_select(cond, self.op_cast(F.tuple_to_array((1.0,)), mstype.float32), self.op_cast(self.beta1, mstype.float32))
        beta2 = self.op_select(cond, self.op_cast(F.tuple_to_array((1.0,)), mstype.float32), self.op_cast(self.beta2, mstype.float32))
        if self.is_group:
            if self.is_group_lr:
                optim_result = self.hyper_map(F.partial(_adam_opt, self.beta1, self.beta2, self.eps),
                                              lr, self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
            else:
                optim_result = self.hyper_map(F.partial(_adam_opt, beta1, beta2, self.eps, lr, overflow),
                                              self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
        else:
            optim_result = self.hyper_map(F.partial(_adam_opt, self.beta1, self.beta2, self.eps, lr, self.weight_decay),
                                          self.parameters, self.moments1, self.moments2,
                                          gradients, self.decay_flags, self.optim_filter)
        if self.use_parallel:
            self.broadcast_params(optim_result)
        return optim_result

class AdamWeightDecayOp(Optimizer):
    """
    Implements the Adam algorithm to fix the weight decay. It is a complete operator, not a combination of other ops.

    Note:
        When separating parameter groups, the weight decay in each group will be applied on the parameters if the
        weight decay is positive. When not separating parameter groups, the `weight_decay` in the API will be applied
        on the parameters without 'beta' or 'gamma' in their names if `weight_decay` is positive.

        To improve parameter groups performance, the customized order of parameters can be supported.

    Args:
        params (Union[list[Parameter], list[dict]]): When the `params` is a list of `Parameter` which will be updated,
            the element in `params` must be class `Parameter`. When the `params` is a list of `dict`, the "params",
            "lr", "weight_decay" and "order_params" are the keys can be parsed.

            - params: Required. The value must be a list of `Parameter`.

            - lr: Optional. If "lr" is in the keys, the value of the corresponding learning rate will be used.
              If not, the `learning_rate` in the API will be used.

            - weight_decay: Optional. If "weight_decay" is in the keys, the value of the corresponding weight decay
              will be used. If not, the `weight_decay` in the API will be used.

            - order_params: Optional. If "order_params" is in the keys, the value must be the order of parameters and
              the order will be followed in the optimizer. There are no other keys in the `dict` and the parameters
              which in the 'order_params' must be in one of group parameters.

        learning_rate (Union[float, Tensor, Iterable, LearningRateSchedule]): A value or a graph for the learning rate.
            When the learning_rate is an Iterable or a Tensor in a 1D dimension, use the dynamic learning rate, then
            the i-th step will take the i-th value as the learning rate. When the learning_rate is LearningRateSchedule,
            use dynamic learning rate, the i-th learning rate will be calculated during the process of training
            according to the formula of LearningRateSchedule. When the learning_rate is a float or a Tensor in a zero
            dimension, use fixed learning rate. Other cases are not supported. The float learning rate must be
            equal to or greater than 0. If the type of `learning_rate` is int, it will be converted to float.
            Default: 1e-3.
        beta1 (float): The exponential decay rate for the 1st moment estimations. Default: 0.9.
            Should be in range (0.0, 1.0).
        beta2 (float): The exponential decay rate for the 2nd moment estimations. Default: 0.999.
            Should be in range (0.0, 1.0).
        eps (float): Term added to the denominator to improve numerical stability. Default: 1e-6.
            Should be greater than 0.
        weight_decay (float): Weight decay (L2 penalty). It must be equal to or greater than 0. Default: 0.0.

    Inputs:
        - **gradients** (tuple[Tensor]) - The gradients of `params`, the shape is the same as `params`.

    Outputs:
        tuple[bool], all elements are True.

    Supported Platforms:
        ``GPU``

    Examples:
        >>> net = Net()
        >>> #1) All parameters use the same learning rate and weight decay
        >>> optim = AdamWeightDecayOp(params=net.trainable_params())
        >>>
        >>> #2) Use parameter groups and set different values
        >>> conv_params = list(filter(lambda x: 'conv' in x.name, net.trainable_params()))
        >>> no_conv_params = list(filter(lambda x: 'conv' not in x.name, net.trainable_params()))
        >>> group_params = [{'params': conv_params, 'weight_decay': 0.01},
        ...                 {'params': no_conv_params, 'lr': 0.01},
        ...                 {'order_params': net.trainable_params()}]
        >>> optim = AdamWeightDecayOp(group_params, learning_rate=0.1, weight_decay=0.0)
        >>> # The conv_params's parameters will use default learning rate of 0.1 and weight decay of 0.01.
        >>> # The no_conv_params's parameters will use learning rate of 0.01 and default weight decay of 0.0.
        >>> # The final parameters order in which the optimizer will be followed is the value of 'order_params'.
        >>>
        >>> loss = nn.SoftmaxCrossEntropyWithLogits()
        >>> model = Model(net, loss_fn=loss, optimizer=optim)
   """
    def __init__(self, params, learning_rate=1e-3, beta1=0.9, beta2=0.999, eps=1e-6, weight_decay=0.0):
        super(AdamWeightDecayOp, self).__init__(learning_rate, params, weight_decay)
        _check_param_value(beta1, beta2, eps, self.cls_name)
        self.beta1 = Tensor(np.array([beta1]).astype(np.float32))
        self.beta2 = Tensor(np.array([beta2]).astype(np.float32))
        self.eps = Tensor(np.array([eps]).astype(np.float32))
        self.moments1 = self.parameters.clone(prefix="adam_m", init='zeros')
        self.moments2 = self.parameters.clone(prefix="adam_v", init='zeros')
        self.hyper_map = C.HyperMap()

    def construct(self, gradients):
        """AdamWeightDecayOp"""
        lr = self.get_lr()
        if self.is_group:
            if self.is_group_lr:
                optim_result = self.hyper_map(F.partial(_adam_opt, self.beta1, self.beta2, self.eps),
                                              lr, self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
            else:
                optim_result = self.hyper_map(F.partial(_adam_opt, self.beta1, self.beta2, self.eps, lr),
                                              self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
        else:
            optim_result = self.hyper_map(F.partial(_adam_opt, self.beta1, self.beta2, self.eps, lr, self.weight_decay),
                                          self.parameters, self.moments1, self.moments2,
                                          gradients, self.decay_flags, self.optim_filter)
        if self.use_parallel:
            self.broadcast_params(optim_result)
        return optim_result



"""
Logging util
@Author: penhe@microsoft.com
"""

__all__ = ['get_logger', 'set_logger']
import logging
import os
import pdb

logging.basicConfig(format = '%(asctime)s|%(levelname)s|%(name)s| %(message)s',
                    datefmt = '%m%d%Y %H:%M:%S',
                    level = logging.INFO)
logger=None
def set_logger(name, file_log=None, rank=0, verbose=1):
    global logger
    if not logger:
      logger = logging.getLogger(name)
    else:
      logger.name = name
    
    dirty_handlers = [h for h in logger.handlers]

    if rank >= 0:
      formatter = logging.Formatter(f'%(asctime)s|%(levelname)s|%(name)s|{rank:02}| %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
    else:
      formatter = logging.Formatter(f'%(asctime)s|%(levelname)s|%(name)s| %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
    if file_log:
        fh = logging.FileHandler(file_log)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    # Stdout
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    if verbose > 0:
      ch.setLevel(logging.INFO)
    else:
      ch.setLevel(logging.WARN)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    for h in dirty_handlers:
      logger.removeHandler(h)
    logger.propagate=False
    return logger

def get_logger(name='logging', file_log=None, rank=0, verbose=1):
  global logger
  if not logger:
    logger = set_logger(name, file_log, rank, verbose)
  return logger




# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import OrderedDict

import numpy as np
import os
from shutil import copyfile
from loguru import logger
from MindsporeTrainer.data import ExampleInstance, ExampleSet, _truncate_segments
from MindsporeTrainer.data.example import *
from MindsporeTrainer.task import EvalData, Task
from MindsporeTrainer.task  import register_task
from utils.metrics import *

from mindspore.communication import init, get_rank, get_group_size

from MindsporeTrainer.modeling.modeling_adapter import NetworkWithLoss
from MindsporeTrainer.utils.metrics import BertMetric, MSAucuracy
from MindsporeTrainer.utils.masker import NGramMaskGenerator
from MindsporeTrainer.data.dynamic_dataset import create_dynamic_dataset
from MindsporeTrainer.modeling.tokenizers import BertTokenizer
from MindsporeTrainer.modeling.layers import ClsEvalHead
from ..models.classify import BertForClassify, ClsLoss

@register_task(name="GJC", desc="GJC task")
class GJCTask(Task):
  def __init__(self, data_dir, args=None, **kwargs):
    super().__init__(args, **kwargs)
    self.data_dir = data_dir
    self.args = args
    # self.model_class = resnet50(32, 100)
    self.metric = 'acc'
    self.main_metric = 'acc'
    self.optimizer_name = 'Lamb'
    self.tokenizer = BertTokenizer(self.args.vocab_path)
    self.label_to_id = {l: i for i, l in enumerate(self.get_labels())}
    if args.distribute:
      self.rank_id = get_rank()
      self.rank_size = get_group_size()
    else:
      self.rank_id = 0
      self.rank_size = 1

  def train_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data = self.load_data(os.path.join(self.data_dir, 'train.tsv'), max_seq_len=max_seq_len)
    # data = ExampleSet(data)
    output_columns = ["input_ids", "input_mask", "token_type_id", "labels"]
    return create_dynamic_dataset(data, self.get_feature_fn(), batch_size=batch_size,
                                  output_columns=output_columns, repeat=self.args.num_train_epochs,
                                  num_workers=self.args.data_workers,num_shards=self.rank_size, shard_id=self.rank_id)

  def eval_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data = self.load_data(os.path.join(self.data_dir, 'dev.tsv'), max_seq_len=max_seq_len)
    # data = ExampleSet(data)
    output_columns = ["input_ids", "input_mask", "token_type_id", "labels"]
    return create_dynamic_dataset(data, self.get_feature_fn(), batch_size=batch_size,
                                  output_columns=output_columns, repeat=self.args.num_train_epochs,
                                  num_workers=self.args.data_workers,num_shards=self.rank_size, shard_id=self.rank_id)

  def test_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data = self.load_data(os.path.join(self.data_dir, 'test.tsv'), max_seq_len=max_seq_len)
    # data = ExampleSet(data)
    output_columns = ["input_ids", "input_mask", "token_type_id", "labels"]
    return create_dynamic_dataset(data, self.get_feature_fn(), batch_size=batch_size,
                                  output_columns=output_columns, repeat=self.args.num_train_epochs,
                                  num_workers=self.args.data_workers,num_shards=self.rank_size, shard_id=self.rank_id)

  def _data(self, name, path, type_name = 'dev', ignore_metric=False, max_examples=None, shuffle=False, max_seq_len=512):
    input_src = os.path.join(self.data_dir, path)
    assert os.path.exists(input_src), f"{input_src} doesn't exists"
    data = self.create_dataset(name)

    predict_fn = self.get_predict_fn(data)
    metrics_fn = self.get_metrics()
    return EvalData(name, data,
      metrics_fn = metrics_fn, predict_fn = predict_fn, ignore_metric=ignore_metric, critial_metrics=['accuracy'])

  def get_metrics(self, **kwargs):
    """Calcuate metrics based on prediction results"""
    return {'acc': MSAucuracy()}

  # def get_predict_fn(self, data):
  #   """Calcuate metrics based on prediction results"""
  #   def predict_fn(logits, output_dir, name, prefix, targets=None):
  #     output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
  #     preds = np.argmax(logits, axis=-1)
  #     labels = self.get_labels()
  #     with open(output, 'w', encoding='utf-8') as fs:
  #       fs.write('targets\tpredictions\n')
  #       if targets is not None:
  #         for i,(e, p) in enumerate(zip(targets, preds)):
  #           fs.write(f'{labels[e]}\t{labels[p]}\n')
  #       else:
  #         for i,(e,p) in enumerate(zip(data, preds)):
  #           fs.write(f'{labels[e.label]}\t{labels[p]}\n')
  #   return predict_fn

  def get_eval_fn(self, **kwargs):
    data = kwargs.get('data')
    if data is None:
      data = self.eval_data(**kwargs)
    def run_eval(model, name, prefix):
      '''
      args: 
        model: Model instance
        name: evaluate name
        prefix: prefix of file
      return:
        float, main metric of this task, used to save best metric model
      '''
      # from wywLM.utils.callbacks import EvalResultsCallback, LossMoniter
      # rc = EvalResultsCallback(result_fn='argmax')
      res = model.eval(data, dataset_sink_mode=False)
      # res = res['acc']
      main_metric = res[self.main_metric]
      if self.rank_id == 0:
        output=os.path.join(self.args.output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
        metric_str = '\n'.join([f'{k}: {v:.4f}' for k, v in res.items()])
        metric_str = metric_str + '\n'
        logger.info("====================================")
        logger.info("evaluate result:\n")
        logger.info(metric_str)
        logger.info("====================================")

        with open(output, 'w', encoding='utf-8') as fs:
          fs.write(f"metrics:\n{metric_str}\n")
      return main_metric
    return run_eval

  def get_feature_fn(self, max_seq_len=512, ext_params=None, rng=None, **kwargs):
    tokenizer = self.tokenizer
    def example_to_feature(*example):
      # print(example)
      features = OrderedDict()
      text = example[0].tolist().decode('utf-8')
      # text = [t.decode('utf-8') for t in text]
      label = example[1].tolist()[0]
      # print(f'sampel: {len(example)}, {text}, {label}\n')
      tokens = ['[CLS]'] + list(text) + ['[SEP]']
      num_pads = max_seq_len - len(tokens)
      pad_id = tokenizer._convert_token_to_id(tokenizer.pad_token)
      token_ids = tokenizer.convert_tokens_to_ids(tokens) + [pad_id] * num_pads
      input_mask = [1] * len(tokens) + [0] * num_pads
      token_type_id = [0] * max_seq_len
      features = (token_ids,
                  input_mask,
                  token_type_id,
                  label)
      features = [np.array(f, dtype=np.int32) for f in features]
      return tuple(features)
    return example_to_feature

  def get_labels(self):
    """See base class."""
    return ['儒藏', '诗藏', '佛藏', '易藏', '医藏', '艺藏', '子藏', '集藏', '道藏', '史藏']
  
  def load_data(self, path, max_seq_len=512, max_examples=None, shuffle=False):
    max_len = max_seq_len - 2
    examples=[]
    merged_words = []
    merged_tokens = []
    merged_labels = []
    size = 0
    with open(path, 'r', encoding='utf-8') as f:
      for sent in f:
        label, sent = sent.strip().split('\t')
        label = self.label_to_id[label]

        # tokens = self.tokenizer.tokenize(sent) # for w in sent
        examples.append(ExampleInstance(segments=[sent], label=label, sentence=sent))
        if self.args.debug and len(examples) >= 3000:
          break

      def get_stats(l):
        return f'Max={max(l)}, min={min(l)}, avg={np.mean(l)}'
      ctx_token_size = [sum(len(w) for w in  e.segments[0]) for e in examples]
      logger.info(f'Statistics: {get_stats(ctx_token_size)}, \
                    long={len([t for t in ctx_token_size if t > 500])}/{len(ctx_token_size)}')
      # return examples
    return ExampleSet(examples)


  def get_model(self):
    from MindsporeTrainer.modeling import build_transformer_model, BertModel
    if self.args.fp16:
      compute_type = mstype.float16
    else:
      compute_type = mstype.float32
    model, config = build_transformer_model(self.args.model_config, 
                                            model=BertModel, 
                                            compute_type=compute_type, 
                                            padding_idx=self.tokenizer._convert_token_to_id(self.tokenizer.pad_token))
    model = BertForClassify(model, num_labels=len(self.get_labels()))
    with open(os.path.join(self.args.output_dir, 'config.json'), 'w', encoding='utf-8') as f:
      f.write(config.to_json_string())
    copyfile(self.args.vocab_path, os.path.join(self.args.output_dir, 'vocab.txt'))
    return model

  def get_loss(self, *args, **kwargs):
    return ClsLoss(reduction='mean')

  def get_eval_head(self, *args, **kwargs):
    from MindsporeTrainer.modeling.models import BertEvalHead
    return ClsEvalHead(len(self.get_labels()))

  def get_opt_fn(self, *args, **kwargs):
    # def opt_fn(args, model, **kwargs):
    #   return Momentum(filter(lambda x: x.requires_grad, model.get_parameters()), 0.01, 0.9)
    return None





# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
import random
import time
import numpy as np
from collections import defaultdict
from loguru import logger

from mindspore.communication.management import get_rank
from mindspore.train.model import Model
from mindspore.nn.wrap.loss_scale import DynamicLossScaleUpdateCell
from mindspore.train.callback import CheckpointConfig
from mindspore.train.serialization import load_checkpoint, load_param_into_net
from mindspore.common import set_seed

from MindsporeTrainer.utils import amp
from MindsporeTrainer.utils.metrics import MSAucuracy
from MindsporeTrainer.optims import get_optimizer
from MindsporeTrainer.optims.adam import *
from MindsporeTrainer.modeling.modeling_adapter import *
from MindsporeTrainer.task import Task
from MindsporeTrainer.utils.callbacks import EvalCallBack, StateCallback, ModelCheckpointWithBest
from MindsporeTrainer.utils.checkpoint import load_ckpt

__all__ = ['DistributedTrainer', 'set_random_seed']

def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    set_seed(seed)

class TrainerState:
    def __init__(self, training_steps, name=None, main_metric=''):
        self.__dict__ = defaultdict(float)
        self.loss = 0.0
        self.examples = 0
        self.steps = 0
        self._last_report_step = 0
        self.epochs = 0
        self.next_batch = 0
        self.num_training_steps = training_steps
        self._last_report_time = time.time()
        self.best_steps = 0
        if main_metric in ['loss', 'perplexity']:
            self.neg_metric = True
        else:
            self.neg_metric = False
        self.best_metric = -1e9
        self.name = name
        self.run_id = None
        self.current_loss = 0
        try:
            self.rank = get_rank()
        except:
            self.rank = 0

    def update_step(self, loss, examples, loss_scale, output=None, **kwargs):
        self.examples += examples
        self.loss += loss
        self.current_loss = loss
        self.steps += 1
        self.next_batch += 1
        if not isinstance(loss_scale, (float, int)):
            self.loss_scale = loss_scale.tolist()
        else:
            self.loss_scale = loss_scale
        self.__dict__.update(**kwargs)
    
    def report_state(self):
        if self.steps <= self._last_report_step or self.rank != 0:
            return

        end = time.time()
        start = self._last_report_time
        if self.name is not None:
            tag = f'[{self.name}]'
        else:
            tag = None
        logger.info('{}[{:0.1f}%][{:0.2f}h] Steps={}, loss={:0.5f}, examples={}, loss_scale={:0.1f}, {:0.4f}s'.format(
            tag, 100*self.steps/self.num_training_steps, \
            (self.num_training_steps - self.steps)*(end - start)/((self.steps-self._last_report_step)*3600), 
            self.steps, self.current_loss, self.examples, self.loss_scale, end-start))
        self._last_report_time = end
        self._last_report_step = self.steps


def get_eval_fn(task, data, output_dir, main_metric='acc', custom_metric_fn=None, rank=0):
    def run_eval(model, name, prefix):
        from MindsporeTrainer.utils.callbacks import EvalResultsCallback, LossMoniter
        rc = EvalResultsCallback(result_fn='argmax')
        res = model.eval(data, callbacks = [rc, LossMoniter(10)], dataset_sink_mode=False)
        if rank == 0:
            result = rc.results
            output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
            targets = result.get('label')
            preds = result.get('net_output')

            if custom_metric_fn is not None:
                cms = custom_metric_fn(labels=targets, predicts=preds)
                for m, m_r in cms.items():
                    if m in res:
                        m = m + '_c'
                    res[m] = m_r
            metric_str = '\n'.join([f'{k}: {v:.4f}' for k, v in res.items()])
            metric_str = metric_str + '\n'
            logger.info("====================================")
            logger.info("evaluate result:\n")
            logger.info(metric_str)
            logger.info("====================================")

            labels = task.get_labels()
            with open(output, 'w', encoding='utf-8') as fs:
                fs.write(f"metrics:\n{metric_str}")
                fs.write('targets\tpredictions\n')
                if targets is not None:
                    for i,(e, p) in enumerate(zip(targets, preds)):
                        if p.size > 1:
                            p = ''.join([labels[pp] for pp in p])
                            e = ''.join([labels[pp] for pp in e])
                        else:
                            p = labels[p]
                            e = labels[e]
                        fs.write(f'{e}\t{p}\n')
                else:
                    for i,(e,p) in enumerate(zip(data, preds)):
                        if p.size > 1:
                            p = ''.join([labels[pp] for pp in p])
                            e = ''.join([labels[pp] for pp in e])
                        else:
                            p = labels[p]
                            e = labels[e]
                        fs.write(f'{e}\t{p}\n')
        return res[main_metric]
    return run_eval


def get_pred_fn(task, model, data, output_dir, name, prefix='final', sequence=False, dump=True):
    def run_predict():
        result = model.predict(data, dataset_sink_mode=True)
        if sequence:
            if result.ndim() == 3:
                result = result.argmax(-1)
        else:
            if result.ndim() == 2:
                result = result.argmax(-1)
        result = result.asnumpy()
        if dump:
            output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
            print("====================================")
            print(f"predict finished. writing resutls to {output}")
            print("====================================")

            labels = task.get_labels()
            with open(output, 'w', encoding='utf-8') as fs:
                for i, p in enumerate(result):
                    if p.size > 0:
                        p = ''.join([labels[pp] for pp in P])
                    else:
                        p = labels[p]
                    fs.write(f'{p}\n')
        return result
    return run_predict

class DistributedTrainer:
    def __init__(self, args, task: Task, **kwargs):
        """
        data_fn return tuples (training_dataset, training_steps, train_sampler, batch_scheduler), training_dataset is required
        loss_fn return the loss of current mini-batch and the size of the batch
        optimizer_fn return the created optimizer
        eval_fn return metrics for model selection
        """
        self.__dict__.update(kwargs)
        self.args = args
        self.task = task
        self.initialized = None
        self.before_run()

        self.accumulative_update = 1
        if hasattr(args, 'accumulative_update'):
            self.accumulative_update = args.accumulative_update

        self.output_dir = args.output_dir
        self.trainer_state = TrainerState(args.train_steps, name=args.tag, main_metric=self.task.main_metric)
        self.report_interval = 10

    def before_run(self):
        if not self.initialized:
            self.initialize()
        if self.args.num_train_epochs <= 0:
            self.args.num_train_epochs = 1
        self.init_task()
        self._setup_model()
        if self.args.do_train:
            if self.args.train_steps <= 0:
                self.args.train_steps = self.args.num_train_epochs * self.train_data.get_dataset_size() // self.args.accumulation_steps
                logger.info("train steps: {}".format(self.args.train_steps))
            else:
                self.args.train_steps = self.args.train_steps * self.args.accumulation_steps
                self.args.num_train_epochs = self.args.train_steps // self.train_data.get_dataset_size() + 1
            self.args.train_steps = int(self.args.train_steps)
            self.args.num_train_epochs = int(self.args.num_train_epochs)
            self.args.save_eval_steps = int(self.args.save_eval_steps / self.args.device_num)

    def init_task(self):
        label_list = self.task.get_labels()
        self.model = self.task.get_model()
        logger.info(f'Total parameters: {sum([p.size for p in self.model.get_parameters()])}')

        self.test_data = None
        logger.info(f"    Evaluation batch size = {self.args.eval_batch_size}")
        if self.args.do_predict:
            self.test_data = self.task.test_data()
            logger.info(f"    Prediction batch size = {self.args.predict_batch_size}")

        if self.args.do_train:
            self.train_data = self.task.train_data(batch_size=self.args.train_batch_size, debug=self.args.debug)
        else:
            self.train_data = None
        optimizer_fn = self.task.get_opt_fn()
        self.optimizer_fn = optimizer_fn if optimizer_fn is not None else get_optimizer

        def _loss_fn(trainer, model, batch):
            _,loss = model(**batch)
            batch_size = batch['input_ids'].size(0)
            return loss.mean(), batch_size
        loss = self.task.get_loss()
        if loss is None:
            loss = self.task.get_loss_fn(self.model)
        self.loss_fn = loss if loss is not None else _loss_fn
        self.eval_head = self.task.get_eval_head()

    def initialize(self):
        set_random_seed(self.args.seed)
        self.initialized = True

    def run(self):
        callbacks = []
        if self.args.fp16 and self.args.device_target != 'Ascend':
            self.model = amp.build_train_network(self.model, self.loss_fn, level='O2')
        else:
            self.model = NetworkWithLoss(self.model, self.loss_fn, return_all=True)

        if self.args.do_eval:
            metrics = self.task.get_metrics()
            self.eval_data = self.task.eval_data()
            if metrics is None:
                metrics = {'acc': MSAucuracy()}
        else:
            self.eval_model = None
            metrics = None

        if self.args.do_train:
            self.optimizer, opt_overflow = self.optimizer_fn(self.args, self.model, opt_name=self.task.optimizer_name)

            if self.args.enable_save_ckpt == "true" and self.args.rank % min(8, self.args.device_num) == 0:
                    config_ck = CheckpointConfig(save_checkpoint_steps=self.args.save_eval_steps,
                                                keep_checkpoint_max=self.args.save_checkpoint_num)
                    ckpoint_cb = ModelCheckpointWithBest(self.trainer_state, prefix=self.args.task_name,
                                                        directory=self.args.output_dir, config=config_ck)
                    callbacks.append(ckpoint_cb)
            self.initialized = False
            if self.args.fp16 and self.args.enable_lossscale == "true":
                    update_cell = DynamicLossScaleUpdateCell(loss_scale_value=self.args.loss_scale_value,
                                                            scale_factor=self.args.scale_factor,
                                                            scale_window=self.args.scale_window)
            else:
                update_cell = None
            accumulation_steps = self.args.accumulation_steps
            enable_global_norm = self.args.enable_global_norm
            network = model_adapter(self.model, optimizer=self.optimizer,
                                    scale_update_cell=update_cell,
                                    accumulation_steps=accumulation_steps,
                                    enable_global_norm=enable_global_norm,
                                    opt_overflow=opt_overflow,
                                    gpu_target=self.args.device_target=='GPU')

            summary_dir = os.path.join(self.output_dir, 'tblogger', str(self.args.rank))
            callbacks.append(StateCallback(self.trainer_state, self.optimizer, 
                                           report_steps=self.args.report_interval,
                                           summary_dir=summary_dir, train_steps=self.args.train_steps))
            if self.args.do_eval:
                self.eval_model = ModelForEval(self.model, self.eval_head, fp16=self.args.fp16)

            model = Model(network, eval_network=self.eval_model, metrics=metrics)
            if self.args.thor:
                from mindspore.train.train_thor import ConvertModelUtils
                model = ConvertModelUtils().convert_to_thor_model(model, network=network, optimizer=self.optimizer)
            if len(self.args.load_checkpoint_path) > 0:
                load_ckpt(network, self.args.load_checkpoint_path, self.args.restore_by_prefix, self.args.prefix)
            if self.args.do_eval:
                main_metric = self.task.main_metric
                eval_fn = self.task.get_eval_fn(data=self.eval_data, output_dir=self.args.output_dir, rank=self.args.rank)
                if eval_fn is None:
                    eval_fn = get_eval_fn(self.task, self.eval_data, self.args.output_dir, 
                                          rank=self.args.rank, main_metric=main_metric)
                eval_callback = EvalCallBack(model, self.trainer_state, self.eval_data, 0, eval_fn, self.args.train_steps,
                                             eval_steps=self.args.save_eval_steps, rank=self.args.local_rank)
                callbacks = [eval_callback] + callbacks
            cb = self.task.get_callbacks()
            if cb is not None:
                callbacks.append(cb)
            model.train(self.args.num_train_epochs, 
                        self.train_data, 
                        callbacks=callbacks, 
                        dataset_sink_mode=self.args.enable_data_sink,
                        sink_size=self.args.data_sink_steps
                        )

        elif self.args.do_predict:
            self.model = ModelForEval(self.model.net, self.eval_head)
            if len(self.args.load_checkpoint_path) > 0:
                load_ckpt(self.model, self.args.load_checkpoint_path, self.args.restore_by_prefix, self.args.prefix)
            model = Model(self.model)
            pred_fn = self.task.get_pred_fn(data=self.eval_data, output_dir=self.args.output_dir, rank=self.args.local_rank)
            if pred_fn is None:
                pred_fn = get_pred_fn()
            result = pred_fn(self.task, model, self.test_data, self.args.output_dir, 'test')
        elif self.args.do_eval and not self.args.do_train:
            self.model = ModelForEval(self.model, self.eval_head)
            if len(self.args.load_checkpoint_path) > 0:
                load_ckpt(self.model, self.args.load_checkpoint_path, self.args.restore_by_prefix, self.args.prefix)
            model = Model(self.model, eval_network=self.model, metrics=metrics)
            main_metric = self.task.main_metric
            metric = self.task.metric
            eval_fn = self.task.get_eval_fn(data=self.eval_data, output_dir=self.args.output_dir, rank=self.args.local_rank)
            if eval_fn is None:
                eval_fn = get_eval_fn(self.task, self.eval_data, self.args.output_dir, 
                                      rank=self.args.local_rank, main_metric=main_metric)
            eval_fn(model, 'eval', 'final')

    def _setup_model(self):
        if len(self.args.load_checkpoint_path) > 0:
                param_dict = load_checkpoint(self.args.load_checkpoint_path)
                load_param_into_net(self.model, param_dict)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
NOTE:
    Transformer Networks.
    This is an experimental interface that is subject to change or deletion.
"""
from .transformer import AttentionMask, VocabEmbedding, MultiHeadAttention, FeedForward, TransformerEncoder, \
    TransformerDecoder, TransformerEncoderLayer, TransformerDecoderLayer, Transformer, TransformerOpParallelConfig, \
    EmbeddingOpParallelConfig, TransformerRecomputeConfig
from .moe import MoEConfig
from .layers import FixedSparseAttention
from .loss import CrossEntropyLoss
from .op_parallel_config import OpParallelConfig

__all__ = []
__all__.extend(transformer.__all__)
__all__.extend(loss.__all__)
__all__.extend(op_parallel_config.__all__)
__all__.extend(layers.__all__)
__all__.extend(moe.__all__)



"""
Logging util
@Author: penhe@microsoft.com
"""

""" Utils for torch jit tracing customer operators/functions
"""
import os

def traceable(cls):
  """ Decorator over customer functions
      There is an issue for tracing customer python torch Function, using this decorator to work around it.
      e.g.
      @traceable
      class MyOp(torch.autograd.Function):
      xxx
  """

  class _Function(object):
    @staticmethod
    def apply(*args):
      jit_trace = (os.getenv('JIT_TRACE', 'False').lower() == 'true')
      if jit_trace:
        return cls.forward(_Function, *args)
      else:
        return cls.apply(*args)

    @staticmethod
    def save_for_backward(*args):
      pass

  _Function.__name__ = cls.__name__
  _Function.__doc__ = cls.__doc__
  return _Function

class TraceMode():
  """ Trace context used when tracing modules contains customer operators/Functions
  """
  def __enter__(self):
    os.environ['JIT_TRACE'] = 'True'
    return self

  def __exit__(self, exp_value, exp_type, trace):
    del os.environ['JIT_TRACE']



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================


import math
import copy
from functools import lru_cache
from turtle import position
import numpy as np
import mindspore.common.dtype as mstype
import mindspore.nn as nn
import mindspore.ops.functional as F
from mindspore.common.initializer import TruncatedNormal, initializer
from mindspore.ops import operations as P
from mindspore.ops import composite as C
import mindspore.ops as O
from mindspore.common.tensor import Tensor
from mindspore.common.parameter import Parameter

from .layers import *


class Transformer(nn.Cell):
    """模型基类
    """
    def __init__(
        self,
        vocab_size,  # 词表大小
        hidden_size,  # 编码维度
        num_hidden_layers,  # Transformer总层数
        num_attention_heads,  # Attention的头数
        intermediate_size,  # FeedForward的隐层维度
        hidden_act,  # FeedForward隐层的激活函数
        max_position_embeddings = 512,
        dropout_rate=None,  # Dropout比例
        hidden_dropout_prob=0.1,
        initializer_range=0.02,
        attention_probs_dropout_prob=None,  # Attention矩阵的Dropout比例
        embedding_size=None,  # 是否指定embedding_size
        attention_head_size=None,  # Attention中V的head_size
        attention_key_size=None,  # Attention中Q,K的head_size
        sequence_length=None,  # 是否固定序列长度
        residual_attention_scores=False,  # Attention矩阵加残差
        type_vocab_size = 1,
        layer_norm_eps=1e-05,
        compute_type=mstype.float32,
        prefix=None,  # 层名前缀
        name=None,  # 模型名称
        config=None,
        **kwargs
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.attention_head_size = attention_head_size or hidden_size // num_attention_heads
        self.attention_key_size = attention_key_size or self.attention_head_size
        self.intermediate_size = intermediate_size
        self.initializer_range = initializer_range
        self.hidden_dropout_prob = hidden_dropout_prob
        self.dropout_rate = dropout_rate or 0
        self.attention_probs_dropout_prob = attention_probs_dropout_prob or 0
        self.hidden_act = hidden_act
        self.embedding_size = embedding_size or hidden_size
        self.sequence_length = sequence_length
        self.max_position_embeddings = max_position_embeddings
        self.type_vocab_size = type_vocab_size
        self.layer_norm_eps = layer_norm_eps
        self.attention_bias = None
        self.position_bias = None
        self.attention_scores = None
        self.residual_attention_scores = residual_attention_scores
        self.prefix = prefix or ''
        self.name = name
        self.compute_type = compute_type
        self.config = config
        self.embeddings = None
        self.encoder = None
        self.decoder = None
        self.output = None

    def build_layers(self, **kwargs):
        raise NotImplementedError

    def get_attention_bias(self, **kwargs):
        '''
        used to generate complex attention mask, UNILM etc.
        '''
        raise self.attention_bias

    def get_position_bias(self, **kwargs):
        raise self.position_bias

    def get_attention_mask(self, mask):
        # input mask from input index sequence
        if mask.ndim == 2: # (batch_size, seq_len)
            seq_length = F.shape(mask)[1]
            mask = F.cast(F.reshape(mask, (-1, 1, seq_length)), mstype.float32)
        return mask

    def parameter_name_mapping(self):
        return {}

    def construct(self, **kwargs):
        raise NotImplementedError


class BertModel(Transformer):
    """
    Bidirectional Encoder Representations from Transformers.

    Args:
        config (Class): Configuration for BertModel.
        is_training (bool): True for training mode. False for eval mode.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
    """
    def __init__(self,
                 use_relative_positions=False,
                 gradient_checkpointing = False,
                 tokenizer = 'BertTokenizer',
                 use_pos=True,
                 config=None,
                 use_one_hot_embeddings=False,
                 **kwargs):
        super(BertModel, self).__init__(**kwargs)
        config = copy.deepcopy(config)

        self.last_idx = self.num_hidden_layers - 1
        # output_embedding_shape = [-1, self.seq_length, self.embedding_size]

        self.embeddings = Embeddings(use_relative_positions=use_relative_positions, **kwargs)

        self.encoder = Encoder(
            hidden_size=self.hidden_size,
            num_attention_heads=self.num_attention_heads,
            num_hidden_layers=self.num_hidden_layers,
            intermediate_size=self.intermediate_size,
            attention_probs_dropout_prob=self.attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=self.initializer_range,
            hidden_dropout_prob=self.hidden_dropout_prob,
            use_relative_positions=use_relative_positions,
            hidden_act=self.hidden_act,
            compute_type=self.compute_type,
            return_all_encoders=True,
            max_position_embeddings=self.max_position_embeddings,
            config=config)

        self.cast = P.Cast()
        self.dtype = self.compute_type
        self.cast_compute_type = SaturateCast(dst_type=self.compute_type)
        self.slice = P.StridedSlice()

        self.squeeze_1 = P.Squeeze(axis=1)
        self.pooler = nn.Dense(self.hidden_size, self.hidden_size,
                              activation="tanh",
                              weight_init=TruncatedNormal(self.initializer_range)).to_float(self.compute_type)
        self._create_attention_mask_from_input_mask = CreateAttentionMaskFromInputMask()

    def construct(self, input_ids, token_type_ids, input_mask):
        """Bidirectional Encoder Representations from Transformers."""
        # embedding
        embedding_tables = self.embeddings.word_embeddings.embedding_table
        word_embeddings = self.embeddings(input_ids, token_type_ids)['embeddings']

        # attention mask [batch_size, seq_length, seq_length]
        attention_mask = self._create_attention_mask_from_input_mask(input_mask)

        # bert encoder
        encoder_output = self.encoder(self.cast_compute_type(word_embeddings),
                                           attention_mask)['hidden_states']

        sequence_output = self.cast(encoder_output[self.last_idx], self.dtype)

        # pooler
        batch_size = P.Shape()(input_ids)[0]
        sequence_slice = self.slice(sequence_output,
                                    (0, 0, 0),
                                    (batch_size, 1, self.hidden_size),
                                    (1, 1, 1))
        first_token = self.squeeze_1(sequence_slice)
        pooled_output = self.pooler(first_token)
        pooled_output = self.cast(pooled_output, self.dtype)

        return sequence_output, pooled_output, embedding_tables


class BertPreTraining(nn.Cell):
    """
    Bert pretraining network.

    Args:
        config (BertConfig): The config of BertModel.
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings.

    Returns:
        Tensor, prediction_scores, seq_relationship_score.
    """

    def __init__(self, **config):
        super(BertPreTraining, self).__init__()
        self.bert = BertModel(**config)
        self.cls1 = GetMaskedLMOutput(**config)
        self.cls2 = GetNextSentenceOutput(**config)

    def construct(self, *sample):
        '''
            sample: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                      "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
        '''
        input_ids, input_mask, token_type_id = sample[: 3]
        masked_lm_positions = sample[4]
        sequence_output, pooled_output, embedding_table = \
            self.bert(input_ids, token_type_id, input_mask)
        prediction_scores = self.cls1(sequence_output,
                                      embedding_table,
                                      masked_lm_positions)
        seq_relationship_score = self.cls2(pooled_output)
        return prediction_scores, seq_relationship_score


class Deberta(Transformer):
    """
    """
    def __init__(self,
                 config=None,
                 use_one_hot_embeddings=False,
                 **kwargs):
        super(Deberta, self).__init__(**kwargs)
        config = copy.deepcopy(config)

        self.last_idx = self.num_hidden_layers - 1
        # output_embedding_shape = [-1, self.seq_length, self.embedding_size]

        self.embeddings = Embeddings(**kwargs)

        self.encoder = Encoder(
            hidden_size=self.hidden_size,
            num_attention_heads=self.num_attention_heads,
            num_hidden_layers=self.num_hidden_layers,
            intermediate_size=self.intermediate_size,
            attention_probs_dropout_prob=self.attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=self.initializer_range,
            hidden_dropout_prob=self.hidden_dropout_prob,
            use_relative_positions=kwargs.get('relative_attention', False),
            hidden_act=self.hidden_act,
            attention=DisentangledSelfAttention(self.__dict__, config=config),
            compute_type=self.compute_type,
            return_all_encoders=True,
            max_position_embeddings=self.max_position_embeddings,
            config=config)
        self.cast = P.Cast()
        self.dtype = self.compute_type
        self.cast_compute_type = SaturateCast(dst_type=self.compute_type)
        self.slice = P.StridedSlice()

        self.squeeze_1 = P.Squeeze(axis=1)
        self.pooler = nn.Dense(self.hidden_size, self.hidden_size,
                              activation="tanh",
                              weight_init=TruncatedNormal(self.initializer_range)).to_float(self.compute_type)
        self._create_attention_mask_from_input_mask = CreateAttentionMaskFromInputMask()

    def construct(self, input_ids, token_type_ids, input_mask, output_all_encoded_layers=True):
        """Bidirectional Encoder Representations from Transformers."""
        # embedding
        embedding_tables = self.embeddings.word_embeddings.embedding_table

        embeddings = self.embeddings(input_ids, token_type_ids)
        word_embeddings = embeddings['embeddings']
        position_embeddings = embeddings['position_embeddings']

        # attention mask [batch_size, seq_length, seq_length]
        attention_mask = self._create_attention_mask_from_input_mask(input_mask)

        # bert encoder
        encoder_output = self.encoder(self.cast_compute_type(word_embeddings),
                                      attention_mask, output_all_encoded_layers=output_all_encoded_layers)['hidden_states']
        # if output_all_encoded_layers:
        #     sequence_output = [self.cast(eo, self.dtype) for eo in encoder_output]
        # else:
        #     sequence_output = [self.cast(encoder_output[self.last_idx], self.dtype)]

        # pooler
        # encoder_output = tuple(encoder_output)
        batch_size = P.Shape()(input_ids)[0]
        sequence_slice = self.slice(encoder_output[-1],
                                    (0, 0, 0),
                                    (batch_size, 1, self.hidden_size),
                                    (1, 1, 1))
        first_token = self.squeeze_1(sequence_slice)
        pooled_output = self.pooler(first_token)
        pooled_output = self.cast(pooled_output, self.dtype)

        return encoder_output, pooled_output, embedding_tables, position_embeddings


class DebertaPreTraining(nn.Cell):
    """
    DeBerta pretraining network.

    Args:
        config (BertConfig): The config of BertModel.
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings.

    Returns:
        Tensor, prediction_scores, seq_relationship_score.
    """

    def __init__(self, **config):
        super(DebertaPreTraining, self).__init__()
        self.config = config.get('config')
        self.bert = Deberta(**config)
        self.cls1 = GetMaskedLMOutput(**config)
        self.lm_predictions = EnhancedMaskDecoder(**config)
        self.pooled_loss = Tensor(np.zeros((1, 2)))

    def construct(self, *sample):
        '''
            sample: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                      "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
        '''
        input_ids, input_mask, token_type_id = sample[: 3]
        masked_lm_positions, masked_lm_ids, masked_lm_weights = sample[4: 7]
        sequence_output, pooled_output, embedding_table, position_embeddings = \
            self.bert(input_ids, token_type_id, input_mask)
        z_states = position_embeddings
        pred_score = self.lm_predictions(sequence_output, z_states, input_mask, self.bert.encoder.layer[-1],
                                         self.bert.encoder.rel_embeddings.embedding_table,
                                         self.bert.encoder.relative_attention, self.bert.encoder.norm_rel_ebd, self.bert.encoder.LayerNorm)
        prediction_scores = self.cls1(pred_score,
                                      embedding_table,
                                      masked_lm_positions)
        return prediction_scores, self.pooled_loss.repeat(pred_score.shape[0], 0)


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
The basic layer of the Transformer Networks. This is an experimental interface that is subject to
change or deletion.
"""
from functools import wraps, partial
import inspect
import math
import numpy as np
from mindspore import nn, context
from mindspore.common.parameter import Parameter
from mindspore.common.initializer import initializer, Tensor
import mindspore.common.dtype as mstype
from mindspore.common.seed import _get_graph_seed
from mindspore.ops import operations as P
from mindspore._extends import cell_attr_register
from mindspore.nn.cell import Cell
from mindspore.nn.layer.activation import get_activation
from mindspore.ops import functional as F
from mindspore._checkparam import Validator
from mindspore.ops.primitive import constexpr
from .op_parallel_config import default_dpmp_config, OpParallelConfig

__all__ = [
    "FixedSparseAttention"
]


def _args_type_validator_check(*type_args, **type_kwargs):
    """Check whether input data type is correct."""

    def type_check(func):
        sig = inspect.signature(func)
        bound_types = sig.bind_partial(*type_args, **type_kwargs).arguments

        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal bound_types
            bound_values = sig.bind(*args, **kwargs)

            argument_dict = bound_values.arguments
            if "kwargs" in bound_types:
                bound_types = bound_types["kwargs"]
            if "kwargs" in argument_dict:
                argument_dict = argument_dict["kwargs"]
            for name, value in argument_dict.items():
                if name in bound_types:
                    bound_types[name](value, name)
            return func(*args, **kwargs)

        return wrapper

    return type_check


def _valid_type_checks(types, class_name):
    # types should be a list of types, this function check if the type is in the valid dtypes
    def validator_check_func(value, name):
        # The args of Validator.check_type_name is (arg_name, arg_type, valid_types, prim_name)
        # as the input of _args_type_validator_check is fixed, so we need to manually change the input order
        partial_check = partial(Validator.check_type_name, valid_types=types, prim_name=class_name)
        return partial_check(name, type(value))

    return validator_check_func


def _valid_value_checks(types, class_name):
    # the value should be a list of types, this function check if the value is in the valid dtypes
    def validator_check_func(value, name):
        # The args of Validator.check_type_name is (arg_name, arg_type, valid_types, prim_name)
        # as the input of _args_type_validator_check is fixed, so we need to manually change the input order
        partial_check = partial(Validator.check_type_name, valid_types=types, prim_name=class_name)
        return partial_check(name, value)

    return validator_check_func


class _LayerInputCheck:
    """
       A input check class for the inputs of the transformer model.
    """
    @staticmethod
    def check_shape_length(input_shape, param_name, func_name, target_len):
        """
        Check the input shape's length is equal to the expected shape
        :param input_shape(list): a list of the tensor shapes.
        :param param_name(str): the name of the checked parameter.
        :param func_name(str): the name of the function.
        :param target_len: the expected length of the shape.
        :return:
        """
        if not isinstance(target_len, list):
            target_len = [target_len]
        matched = False
        for item in target_len:
            if len(input_shape) == item:
                matched = True
        if not matched:
            raise ValueError(f"{func_name} {param_name} shape length should be one of {target_len} dimension, "
                             f"but got shape {input_shape}")
        return True

    @staticmethod
    def check_shape_equal(input_shape, param_name, func_name, target_shape):
        """
        Check the input shape's is equal to the expected shape
        :param input_shape(list): a list of the tensor shapes.
        :param param_name(str): the name of the checked parameter.
        :param func_name(str): the name of the function.
        :param target_shape: the expected shape.
        :return:
        """
        if not isinstance(target_shape[0], list):
            target_shape = [target_shape]
        if isinstance(input_shape, tuple):
            input_shape = list(input_shape)
        _LayerInputCheck.check_shape_length(input_shape, param_name, func_name,
                                            [len(item) for item in target_shape])
        matched = False
        for item in target_shape:
            if item == input_shape:
                matched = True
                break

        if not matched:
            raise ValueError(f"{func_name} {param_name} shape should be one of {target_shape},"
                             f"but got {input_shape}")
        return True

    @staticmethod
    def check_shape_value_on_axis(input_shape, dim, param_name, cls_name, target_value):
        """ Check whether the input_shape[dim] is equal to target value"""
        if input_shape[dim] != target_value:
            raise ValueError(f"{cls_name} {param_name} at {dim} shape should be {target_value},"
                             f"but got {input_shape[dim]}")
        return True



@constexpr
def _check_past_none_input_none(use_past, param_name, func_name, default_value, is_tensor, is_default):
    """ If the past is True, check whether the inputs is None"""
    if not use_past:
        if is_tensor:
            raise TypeError(f"{func_name} {param_name} should be {default_value}, if use_pat is False, but found "
                            f"a tensor")
        if not is_default:
            raise TypeError(f"{func_name} {param_name} should be {default_value}, if use_pat is False.")
    else:
        if not is_tensor:
            raise TypeError(f"{func_name} {param_name} should be tensor, if use_pat is True")
    return True



@constexpr
def _check_input_dtype(input_dtype, param_name, allow_dtypes, cls_name):
    Validator.check_type_name(param_name, input_dtype, allow_dtypes, cls_name)


@constexpr
def _check_input_shape(input_shape, param_name, func_name, target_len):
    # check the input length
    _LayerInputCheck.check_shape_length(input_shape, param_name, func_name, target_len)


@constexpr
def _check_shape_equal(input_shape, param_name, func_name, target_shape):
    # check the input length
    _LayerInputCheck.check_shape_equal(input_shape, param_name, func_name, target_shape)


@constexpr
def _check_input_shape_value(input_shape, dim, param_name, cls_name, target_value):
    _LayerInputCheck.check_shape_value_on_axis(input_shape, dim, param_name, cls_name, target_value)


class _LayerNorm(Cell):
    r"""
        A self-defined layer norm operation using reduce sum and reduce mean

        Args:
            normalized_shape (tuple): The shape of the input tensor
            eps (float): The epsilon value of the denominator. Default 1e-5.
            param_init_type: The param init type.
        Inputs:
            - **x** (Tensor) - Tensor of shape :math:`(batch, seq\_length, hidden\_size)`.

        Outputs:
            Tensor of shape :math:`(batch, seq_length, hidden_size)`.
    """

    def __init__(self, normalized_shape, eps=1e-5, param_init_type=mstype.float32):
        super(_LayerNorm, self).__init__()
        if param_init_type not in [mstype.float32, mstype.float16]:
            raise TypeError("The type of parameter 'param_init_type' should in [float32, float16], "
                            "but got the type : {}.".format(type(param_init_type)))
        if normalized_shape[0] <= 1024:
            self.layer_norm = P.LayerNorm(begin_norm_axis=-1,
                                          begin_params_axis=-1,
                                          epsilon=eps)
        self.is_self_defined = normalized_shape[0] > 1024
        self.gamma = Parameter(initializer('ones', normalized_shape, param_init_type), name="gamma",
                               parallel_optimizer=False)
        self.beta = Parameter(initializer('zeros', normalized_shape, param_init_type), name="beta",
                              parallel_optimizer=False)
        self.mean = P.ReduceMean(keep_dims=True)
        self.square = P.Square()
        self.sqrt = P.Sqrt()
        self.sub1 = P.Sub()
        self.sub2 = P.Sub()
        self.add = P.Add()
        self.eps = eps
        self.mul = P.Mul()
        self.add2 = P.Add()
        self.real_div = P.RealDiv()

    def construct(self, x):
        r"""
          x : batch x seq_length x hidden_size
        """
        if self.is_self_defined:
            mean = self.mean(x, -1)
            diff = self.sub1(x, mean)
            variance = self.mean(self.square(diff), -1)
            variance_eps = self.sqrt(self.add(variance, self.eps))
            output = self.real_div(diff, variance_eps)
            output = self.add2(self.mul(output, self.gamma), self.beta)
        else:
            output, _, _ = self.layer_norm(x, self.gamma, self.beta)
        return output

    def shard(self, strategy):
        r"""
        Set the shard for the layer norm. the strategy size should be equal to the inputs.

        Note:
            It is valid only in semi auto parallel or auto parallel mode.
            In other parallel modes, strategies set here will be ignored.

        Args:
            strategy (tuple): The strategy for the dropout. Should be the same shape as the inputs.
        Examples:
            >>> import mindspore
            >>> net = mindspore.parallel.nn.transformer.LayerNorm(normalized_shape=(1024, 10))
            >>> net.shard(((10, 2, 1),))
        """
        if self.is_self_defined:
            self.mean.shard(strategy)
            self.square.shard(strategy)
            self.sqrt.shard(strategy)
            self.sub1.shard((strategy[0], strategy[0]))
            self.sub2.shard((strategy[0], strategy[0]))
            self.add.shard((strategy[0], ()))
            self.mul.shard((strategy[0], (1,)))
            self.add2.shard((strategy[0], (1,)))
            self.real_div.shard((strategy[0], strategy[0]))
        else:
            self.layer_norm.shard((strategy[0], (1,), (1,)))

        return self


class _Linear(Cell):
    r"""
    The dense connected layer. Once the parallel mode is enabled, the input shape should be
    3-D tensor.

    Applies dense connected layer for the input. This layer implements the operation as:

    .. math::
        \text{outputs} = \text{activation}(\text{X} * \text{kernel} + \text{bias}),

    where :math:`X` is the input tensors, :math:`\text{activation}` is the activation function passed as the activation
    argument (if passed in), :math:`\text{kernel}` is a weight matrix with the same
    data type as the :math:`X` created by the layer, and :math:`\text{bias}` is a bias vector
    with the same data type as the :math:`X` created by the layer (only if has_bias is True).

    Args:
        in_channels (int): The number of channels in the input space.
        out_channels (int): The number of channels in the output space.
        weight_init (Union[Tensor, str, Initializer, numbers.Number]): The trainable weight_init parameter. The dtype
            is same as `x`. The values of str refer to the function `initializer`. Default: 'normal'.
        bias_init (Union[Tensor, str, Initializer, numbers.Number]): The trainable bias_init parameter. The dtype is
            same as `x`. The values of str refer to the function `initializer`. Default: 'zeros'.
        has_bias (bool): Specifies whether the layer uses a bias vector. Default: True.
        activation (str): activate function applied to the output of the fully connected layer,
            eg. 'ReLU'.Default: None.
        expert_num (int): The number of experts used in this Linear. Here, for the case expert_num > 1, BatchMatMul is
            used and the first dimension in BatchMatMul indicate expert_num. Default: 1.
        outer_batch (int): The replication number of experts. The replication is effective only when MoE is applied.
            Default: 1.
        compute_dtype (dtype.Number): The computation type. Default: mstype.float16
    Inputs:
        - **x** (Tensor) - Tensor of shape :math:`(*, in\_channels)`. The `in_channels` in `Args` should be equal
          to :math:`in\_channels` in `Inputs`.

    Outputs:
        Tensor of shape :math:`(*, out\_channels)`.

    Raises:
        TypeError: If `in_channels` or `out_channels` is not an int.
        TypeError: If `has_bias` is not a bool.
        TypeError: If `activation` is not one of str, Cell, Primitive, None.
        ValueError: If length of shape of `weight_init` is not equal to 2 or shape[0] of `weight_init`
                    is not equal to `out_channels` or shape[1] of `weight_init` is not equal to `in_channels`.
        ValueError: If length of shape of `bias_init` is not equal to 1
                    or shape[0] of `bias_init` is not equal to `out_channels`.

    Supported Platforms:
        ``Ascend`` ``GPU``
    """

    @cell_attr_register
    @_args_type_validator_check(in_channels=Validator.check_positive_int,
                                out_channels=Validator.check_positive_int,
                                has_bias=Validator.check_bool,
                                activation=_valid_type_checks([type(None), str], "Linear"),
                                transpose_b=Validator.check_bool,
                                expert_num=Validator.check_positive_int,
                                outer_batch=Validator.check_positive_int,
                                param_init_type=_valid_value_checks([mstype.float32, mstype.float16],
                                                                    "Linear"),
                                compute_dtype=_valid_value_checks([mstype.float32, mstype.float16],
                                                                  "Linear"))
    def __init__(self,
                 in_channels,
                 out_channels,
                 weight_init='normal',
                 bias_init='zeros',
                 has_bias=True,
                 activation=None,
                 transpose_b=True,
                 expert_num=1,
                 outer_batch=1,
                 param_init_type=mstype.float32,
                 compute_dtype=mstype.float16):
        super(_Linear, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        if isinstance(weight_init, Tensor) and (weight_init.ndim != 2 or weight_init.shape[0] != out_channels or \
                    weight_init.shape[1] != in_channels):
            raise ValueError("The shape of parameter 'weight_init' is error, please check shape of 'weight_init'.")
        weight_shape = [out_channels, in_channels] if transpose_b else [in_channels, out_channels]
        self.expert_num = expert_num
        self.outer_batch = outer_batch
        if self.expert_num > 1:
            self.expert_flag = True
            self.weight = Parameter(initializer(weight_init, [self.expert_num] + weight_shape, param_init_type),
                                    name="weight")
            self.matmul = P.BatchMatMul(transpose_b=transpose_b)
        else:
            self.expert_flag = False
            self.weight = Parameter(initializer(weight_init, weight_shape, param_init_type), name="weight")
            self.matmul = P.MatMul(transpose_b=transpose_b)
        self.bias = None
        self.has_bias = has_bias
        if self.has_bias:
            if isinstance(bias_init, Tensor) and (bias_init.ndim != 1 or bias_init.shape[0] != out_channels):
                raise ValueError("The shape of parameter 'bias_init' is error, please check shape of 'bias_init'.")
            self.bias = Parameter(initializer(bias_init, [out_channels], param_init_type), name="bias")
            self.bias_add = P.Add()
        self.act_name = activation
        self.activation = get_activation(activation) if isinstance(activation, str) else activation
        self.activation_flag = self.activation is not None
        self.dtype = compute_dtype
        self.cast = P.Cast()

    def construct(self, x):
        out_shape = P.Shape()(x)[:-1] + (self.out_channels,)
        x = P.Reshape()(x, (-1, self.in_channels))
        if self.expert_flag:
            x = P.Reshape()(x, (self.outer_batch, self.expert_num, -1, self.in_channels))
        weight = self.cast(self.weight, self.dtype)
        x = self.matmul(x, weight)
        if self.has_bias:
            x = self.bias_add(x, self.cast(self.bias, self.dtype))
        if self.activation_flag:
            x = self.activation(x)
        output = P.Reshape()(x, out_shape)
        return output

    def shard(self, strategy_matmul, strategy_bias=None, strategy_activation=None):
        r"""
         Set the shard for the linear. the strategy size should be equal to the inputs.

         Note:
            It is valid only in semi auto parallel or auto parallel mode.
            In other parallel modes, strategies set here will be ignored.

         Args:
             strategy_matmul (tuple): The strategy for the matmul. Should be the same shape as the inputs.
             strategy_bias (tuple): The strategy for the bias_add. Should be the same shape as the inputs.
             strategy_activation (tuple): The strategy for the strategy_activation. Should be the same shape as
                the inputs.
         """
        self.matmul.shard(strategy_matmul)
        if self.has_bias:
            self.bias_add.shard(strategy_bias)
        if self.activation_flag:
            # some operations has many primitives, need to manually set the shard
            if self.act_name.lower() == "leakyrelu":
                self.activation.select_op.shard((strategy_activation[0], strategy_activation[0]))
            elif self.act_name.lower() == "logsigmoid":
                self.activation.mul.shard((strategy_activation[0], ()))
                self.activation.exp.shard(strategy_activation)
                self.activation.add.shard((strategy_activation[0], ()))
                self.activation.rec.shard(strategy_activation)
                self.activation.log.shard(strategy_activation)
            elif self.act_name.lower() == "logsoftmax":
                raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                 "or auto parallel mode.")
            else:
                getattr(self.activation, self.act_name).shard(strategy_activation)

        return self


class _Dropout(nn.Cell):
    r"""
        A Dropout Implements with P.DropoutGenMask and  P.DropoutDoMask for parallel training.
    """

    def __init__(self, keep_prob=0.5, dtype=mstype.float32):
        super(_Dropout, self).__init__()
        if keep_prob <= 0 or keep_prob > 1:
            raise ValueError(
                "dropout probability should be a number in range (0, 1], but got {}".format(
                    keep_prob))
        Validator.check_subclass("dtype", dtype, mstype.number_type, self.cls_name)
        Validator.check_value_type('keep_prob', keep_prob, [float], self.cls_name)
        self.keep_prob = keep_prob
        self.is_ascend = context.get_context('device_target') in ["Ascend"]
        if self.is_ascend:
            seed0, seed1 = _get_graph_seed(0, "dropout")
            self.seed0 = seed0
            self.seed1 = seed1
            self.dtype = dtype
            self.get_shape = P.Shape()
            self.dropout_gen_mask = P.DropoutGenMask(Seed0=self.seed0, Seed1=self.seed1)
            self.dropout_do_mask = P.DropoutDoMask()
            self.cast = P.Cast()
        else:
            self.dropout = P.Dropout(keep_prob)

    def construct(self, x):
        r"""
           Input: a tensor
           Returns: a tensor
        """
        if not self.training:
            return x

        if self.keep_prob == 1:
            return x

        if not self.is_ascend:
            out, _ = self.dropout(x)
            return out

        shape = self.get_shape(x)
        dtype = P.DType()(x)
        keep_prob = self.cast(self.keep_prob, dtype)
        output = self.dropout_gen_mask(shape, keep_prob)
        return self.dropout_do_mask(x, output, keep_prob)

    def extend_repr(self):
        return 'keep_prob={}'.format(self.keep_prob)

    def shard(self, strategy):
        if self.is_ascend:
            self.dropout_gen_mask.shard(strategy)
            self.dropout_do_mask.shard(strategy)
        else:
            self.dropout.shard(strategy)


class FixedSparseAttention(nn.Cell):
    """
    Fixed Sparse Attention Layer

    This function contains the sparse attention primitives used in Sparse Transformers (see paper).
    `Generating Long Sequences with Sparse Transformers <https://arxiv.org/abs/1904.10509>`_.
    Specifically, it includes the following:
    1. A faster implementation of normal attention (the upper triangle is not computed, and many operations are fused).
    2. An implementation of "strided" and "fixed" attention, as in the Sparse Transformers paper.

    Args:
        batch_size(int): Number of input batch size.
        num_heads(int): Number of attention heads.
        size_per_head(int): An integer determining embedding size of each attention head,
            only supports 64, 128 for now.
        block_size(int): An integer determining the block size. Current implementation of sparse self-attention
            is based on blocked sparse matrices. In which this parameter defines the size of such blocks,
            Block X Block. Only supports 64 for now.
        seq_length(int): length of input sequence, only supports 1024 for now. Default 1024.
        num_different_global_patterns(int): An integer determining the number of different global attentions layouts.
            While global attention can be fixed by which block/s are representative of
            any local window, since there are multi-heads, each head can use a
            different global representative, only supports 4 for now. Default 4.
        parallel_config(OpParallelConfig): The config of parallel setting, see `OpParallelConfig`.
            Default `default_dpmp_config`, an instance of `OpParallelConfig` with
            default args.

    Inputs:
        - **q** (Tensor) - Tensor query (:class:`mstype.fp16` [batch_size, seq_length, hidden_size]): Sequence of
          queries to query the context.
        - **k** (Tensor) - Tensor key (:class:`mstype.fp16` [batch_size, seq_length, hidden_size]): Sequence of
          queries to query the context.
        - **v** (Tensor) - Tensor value (:class:`mstype.fp16` [batch size, sequence length, Embedding Size]):
          Sequence of queries to query the context.
        - **attention_mask** (Tensor) - Float Tensor the mask of (:class:`mstype.fp32`, :class:`mstype.fp16`
          [batch_size, seq_length, seq_length]): Lower triangular matrix to pass masked information.

    Outputs:
        A Tensor. The output of the attention with shape [batch_size, seq_length, hidden_size]

    Supported Platforms:
        ``Ascend``

    Examples:
        >>> import numpy as np
        >>> from mindspore import dtype as mstype
        >>> from mindspore.nn.transformer import FixedSparseAttention
        >>> from mindspore import Tensor
        >>> model = FixedSparseAttention(batch_size=2,
        ...                              num_heads=8,
        ...                              size_per_head=64,
        ...                              block_size=64)
        >>> q = Tensor(np.ones((2, 1024, 8*64)), mstype.float16)
        >>> k = Tensor(np.ones((2, 1024, 8*64)), mstype.float16)
        >>> v = Tensor(np.ones((2, 1024, 8*64)), mstype.float16)
        >>> attention_mask = Tensor(np.ones((2, 1024, 1024)), mstype.float32)
        >>> output = model(q, k, v, attention_mask)
        >>> print(output.shape)
        (2, 1024, 512)
    """

    @_args_type_validator_check(batch_size=Validator.check_positive_int,
                                num_heads=Validator.check_positive_int,
                                size_per_head=Validator.check_positive_int,
                                block_size=Validator.check_positive_int,
                                seq_length=Validator.check_positive_int,
                                num_different_global_patterns=Validator.check_positive_int,
                                parallel_config=_valid_type_checks([OpParallelConfig], "FixedSparseAttention"))
    def __init__(self,
                 batch_size,
                 num_heads,
                 size_per_head,
                 block_size,
                 seq_length=1024,
                 num_different_global_patterns=4,
                 parallel_config=default_dpmp_config):
        super(FixedSparseAttention, self).__init__()
        dp, mp = parallel_config.data_parallel, parallel_config.model_parallel
        if num_heads % mp != 0:
            raise ValueError(f"The number of heads {num_heads} must be a "
                             f"multiple of parallel_config.model_parallel {mp}.")
        if batch_size % dp != 0:
            raise ValueError(f"The batch_size {batch_size} must be a "
                             f"multiple of parallel_config.data_parallel {parallel_config.data_parallel}.")
        self.seq_length = seq_length
        self.batch_size = batch_size
        self.hidden_size = size_per_head * num_heads
        self.num_heads = num_heads
        self.block_size = block_size
        self.block_num = seq_length // block_size
        self.size_per_head = size_per_head
        self.global_size = seq_length // 4
        self.reshape = P.Reshape()
        self.transpose = P.Transpose().shard(((dp, 1, mp, 1),))
        self.batch_matmul = P.BatchMatMul().shard(((dp, 1, 1, 1), (dp, 1, 1, 1)))
        self.multiply = P.Mul().shard(((dp, 1, 1, 1), (1, 1, 1)))
        self.multiply_data = Tensor([-10000.0], dtype=mstype.float32)
        self.parallel_config = parallel_config
        size_per_head_list = [64, 128]
        if self.seq_length != 1024:
            raise ValueError("For 'FixedSparseAttention', the class variable 'seq_length' must be 1024, "
                             "but got the value : {}.".format(seq_length))
        if self.block_size != 64:
            raise ValueError("For 'FixedSparseAttention', the class variable 'block_size' must be 64, "
                             "but got the value : {}.".format(block_size))
        if num_different_global_patterns != 4:
            raise ValueError("For 'FixedSparseAttention', the class variable 'num_different_global_patterns' "
                             "must be 4, but got the value : {}".format(num_different_global_patterns))
        if self.size_per_head not in size_per_head_list:
            raise ValueError("For 'FixedSparseAttention', the class variable 'size_per_head' only supports {}, "
                             "but got the value : {}.".format(size_per_head_list, self.size_per_head))
        local_ones = np.ones((self.block_size, self.block_size),
                             dtype=np.float16)
        global_mask_original = np.ones((self.seq_length, self.global_size), dtype=np.float16)
        for i in range(self.seq_length):
            for j in range(self.global_size):
                if i // 16 >= (j // 16 + 1) * 4:
                    global_mask_original[i, j] = 0.0

        global_mask_original = -10000 * global_mask_original
        global_mask_fx = global_mask_original.reshape((self.seq_length // 16, 16, self.global_size // 16, 16))
        global_mask = np.transpose(global_mask_fx, (2, 0, 1, 3))
        global_mask = np.repeat(global_mask[np.newaxis, :, :, :, :], self.batch_size, axis=0)
        global_mask = global_mask.reshape((self.batch_size * self.global_size // 16, self.seq_length // 16, 16, 16))
        self.global_mask = Tensor(global_mask, mstype.float32)
        self.local_mask_triangle = Tensor(np.tril(local_ones), mstype.float32)
        self.scale_factor = Tensor((math.sqrt(self.size_per_head)))
        self.matmul_dds = P.MatmulDDS(self.batch_size, self.num_heads).shard(((mp, dp, 1, 1),
                                                                              (mp, dp, 1, 1),
                                                                              (1, dp, 1, 1),
                                                                              (dp, 1, 1, 1)))
        self.matmul_dsd = P.DSDMatmul().shard(((dp, mp, 1, 1, 1, 1, 1), (dp, mp, 1, 1, 1, 1, 1), (dp, mp, 1, 1)))
        self.sub1 = P.Sub().shard(((1,), (dp, 1, 1, 1)))
        self.mul1 = P.Mul().shard(((dp, 1, 1, 1), (1,)))
        self.transpose1 = P.Transpose().shard(((dp, 1, 1, 1),))
        self.transpose2 = P.Transpose().shard(((dp, 1, 1, 1),))
        self.transpose3 = P.Transpose().shard(((dp, mp, 1, 1, 1, 1),))
        self.transpose4 = P.Transpose().shard(((dp, mp, 1, 1),))
        self.div = P.RealDiv().shard(((mp, dp, 1, 1), ()))
        self.slice1 = P.StridedSlice().shard(((dp, 1, 1),))

    def _transpose_inputs(self, q, k, v):
        """
        do reshape and transpose to inputs
        """
        q = self.transpose(
            self.reshape(
                q,
                (-1, 16, self.num_heads * self.size_per_head // 16, 16)),
            (2, 0, 1, 3))
        k = self.transpose(
            self.reshape(
                k, (-1, 16, self.num_heads * self.size_per_head // 16, 16)),
            (2, 0, 1, 3))
        v = self.transpose(
            self.reshape(
                v,
                (-1, 16, self.num_heads * self.size_per_head // 16, 16)),
            (0, 2, 3, 1))

        return q, k, v

    def _generate_attention_mask(self, attention_mask):
        """
        generate global attention mask and local attention mask from origin attention mask
        """
        attention_mask = self.reshape(attention_mask, (-1, self.seq_length, self.seq_length))
        input_mask = self.slice1(attention_mask, (0, self.seq_length - 1, 0),
                                 (self.batch_size, self.seq_length, self.seq_length), (1, 1, 1))
        input_mask = self.reshape(input_mask, (-1, self.seq_length))
        input_shape = P.Shape()(input_mask)  # bs, seq_length
        # bs, block_num, 1, block_size
        local_shape_right = (input_shape[0], self.block_num, 1, self.block_size)
        # bs, block_num, block_size, 1
        local_shape_left = (input_shape[0], self.block_num, self.block_size, 1)
        local_mask_left = self.reshape(input_mask, local_shape_left)
        local_mask_right = self.reshape(input_mask, local_shape_right)
        # bs, block_num, block_size, block_size
        local_attention_mask = self.batch_matmul(local_mask_left, local_mask_right)
        lower_triangle = P.ExpandDims()(self.local_mask_triangle, 0)
        local_attention_mask = self.multiply(local_attention_mask, lower_triangle)
        local_multiplied_out = self.sub1(P.Cast()(F.tuple_to_array((1.0,)), mstype.float32),
                                         P.Cast()(local_attention_mask, mstype.float32))
        local_adder = self.mul1(local_multiplied_out, self.multiply_data)
        local_mask_original = self.transpose1(local_adder, (0, 2, 1, 3))
        local_mask_original = self.reshape(
            local_mask_original,
            (self.batch_size * self.block_size, self.block_num * self.block_size))
        local_mask_fx = self.reshape(
            local_mask_original,
            (self.batch_size * self.block_size // 16, 16,
             self.block_num * self.block_size // 16, 16))
        local_mask = self.transpose2(local_mask_fx, (2, 0, 1, 3))
        global_mask = self.global_mask

        return local_mask, global_mask

    def construct(self, q, k, v, attention_mask):
        _check_shape_equal(F.shape(q), "q", self.cls_name,
                           [self.batch_size, self.seq_length, self.hidden_size])
        _check_input_dtype(F.dtype(q), "q", [mstype.float16], self.cls_name)
        _check_shape_equal(F.shape(k), "k", self.cls_name,
                           [self.batch_size, self.seq_length, self.hidden_size])
        _check_input_dtype(F.dtype(k), "k", [mstype.float16], self.cls_name)
        _check_shape_equal(F.shape(v), "v", self.cls_name,
                           [self.batch_size, self.seq_length, self.hidden_size])
        _check_input_dtype(F.dtype(v), "v", [mstype.float16], self.cls_name)
        _check_shape_equal(F.shape(attention_mask), "attention_mask", self.cls_name,
                           [self.batch_size, self.seq_length, self.seq_length])
        _check_input_dtype(F.dtype(attention_mask), "attention_mask", [mstype.float32, mstype.float16], self.cls_name)

        q, k, v = self._transpose_inputs(q, k, v)
        local_mask, global_mask = self._generate_attention_mask(attention_mask)
        q = self.div(q, F.cast(self.scale_factor, F.dtype(q)))
        k = self.div(k, F.cast(self.scale_factor, F.dtype(k)))
        local_prob, global_prob = self.matmul_dds(q, k, local_mask, global_mask)
        attention = self.matmul_dsd(local_prob, global_prob, v)
        attention_merge = self.transpose3(attention, (0, 1, 3, 4, 2, 5))
        attention_merge = F.reshape(
            attention_merge,
            (-1, self.num_heads, self.seq_length, self.size_per_head))
        attention_merge = self.transpose4(attention_merge, (0, 2, 1, 3))
        attention_merge = F.reshape(
            attention_merge,
            (-1, self.seq_length, self.size_per_head * self.num_heads))

        return attention_merge



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from .example import ExampleInstance, ExampleSet, _truncate_segments
# from .dataloader import SequentialDataLoader
from .dynamic_dataset import *
# from .data_sampler import *
# from .async_data import *
# from .collate import *
